#include <iostream>
#include <ctime>
#include <cfloat>
#include "float.h"
#include <stdlib.h>
#include <string>
#include <math.h>
#include<iostream>
#include<string>
#include<sstream>
#include "anyoption.h"
#include "matrixRow.h"
#include <limits.h>

#define ROWCLUSTERS   10
#define COLCLUSTERS   10
#define MAXITERATIONS 100

#define CREATE(name,size)  float *name= new float[size]; for (unsigned i=0; i<size; i++) name[i]=0;
#define CHECK(name,size)   for (unsigned i=0; i<size; i++) if (name[i]==0.0) {\
emptycluster = true; /*printf("EMPTY %s %i\n",#name, i);*/ }

using namespace std;

unsigned k = ROWCLUSTERS;   // The desired number of row clusters.
unsigned l = COLCLUSTERS;   // The desired number of column clusters.
unsigned m = MAXITERATIONS; // The maximum number of iterations to which cocluster will end.
unsigned n = 0;             // The number of rows in the input matrix.
unsigned nCols = 0;         // The number of columns in the input matrix.
bool verbose = false;
bool dump_all = false;
bool dump_pcap = true;
bool information = false;



bool toss_coin(double probability)
{
  double coin = (1.0)*rand()/(RAND_MAX+1.0);

  //  if (coin < probability)
  //  printf("XX ");

  return (coin < probability);
}




void completion(int value, int frequency,char c)
{
  if (verbose)

    if (((value + 1) % frequency) == 0)
      {
	putchar(c);
      }
  
  return;
}




string itos(int i)	// convert int to string
{
  stringstream s;
  s << i;
  return s.str();
}




char *timestamp(char *buff) {
  
  time_t startTime = time(NULL);
  int len;

  buff = ctime(&startTime);
  len = strlen(buff);
  buff[len-1]='\0';
  
  return buff;

}




/* 

	Following is the accepted format for the data being read:
	The input file is a binary file representing a sparse matrix.
	In this version the sparse matrix is kept uncompressed, following versions
	will adopt matrix compression heuristics.
	In this file a row-based representation of the matrix is contained.
	Each row is represented in the following way:
	<e> <element_1,value_1> <element_2,value_2> ... <element_e,value_e>
	where:
		<e>         is a unsigned integer;
		<element_i> is a unsigned integer;
		<value_i>   is a float.
	
	
	Example.
	If we want to represent the following matrix:
	
	    	| 0.0 2.3 0.3 4.0 | 
	    A=  | 1.2 0.0 0.0 2.0 |
	    	| 0.0 1.5 1.6 0.0 |
	
	Then file will contain
	
	3 <1,2.3><2,0.3><3,4.0>
	2 <0,1.2><3,2.0>
	2 <1,1.5><1,1.6>
	

*/
string   inputRowFile;           // This is the input file from which data will be read.
ifstream *inRow;

/* 

Following is the accepted format for the data being read:
The input file is a binary file representing a sparse matrix.
In this version the sparse matrix is kept uncompressed, following versions
will adopt matrix compression heuristics.

In this file a column-based representation of the matrix is contained.
Each column is represented in the following way:
	<e> <element_1,value_1> <element_2,value_2> ... <element_e,value_e>
where:
<e>         is a unsigned integer;
<element_i> is a unsigned integer;
<value_i>   is a float.


Example.
If we want to represent the following matrix:

    | 0.0 2.3 0.3 4.0 | 
A=  | 1.2 0.0 0.0 2.0 |
    | 0.0 1.5 1.6 0.0 |

Then file will contain

1 <1,1.2>
2 <0,2.3><2,1.5>
2 <1,0.3><2,1.6>
2 <0,4.0><1,2.0>

*/
string   inputColumnFile;           // This is the input file from which data will be read.
ifstream *inColumn;

/*

	The output will be generated according to this format:
	The output file is a text file containing firstly the mapping of rows into clusters
	and secondly the mapping of columns into clusters.
	
	For example let's suppose X={x_1,x_2,x_3,x_4, x_5} and Y={y_1,y_2,y_3,y_4} a
	co-clustering assigning {x_1,x_3} to row-cluster 1, {x_2,x_4,x_5} to row-cluster 2,
	{y_1} to column-cluster 1, {y_3} to column-cluster 2, and {y_2,y_4} to column-cluster 3.
	
	The output file will be the following text file:
	
	<---- Row Clusters ---->
	0
	1
  0
  1
  1
	<---- Column Clusters ---->
	0
  2
  1
  2




  If requested, the file .pcap will store the computed values of the coclustered matrix, in this format:

  one line per row cluster
  one value per column 
  space-separated

	
*/


string    outputFile;			 // This is the file to which output will be written.


void readCommandLine (int argc, char* argv[])
{
	/* 1. CREATE AN OBJECT */
	AnyOption *opt = new AnyOption();

	/* 2. SET PREFERENCES  */
	//opt->noPOSIX(); /* do not check for POSIX style character options */
	//opt->setVerbose(); /* print warnings about unknown options */
	//	opt->noUsage(); /* stop printing Usage */

	/* 3. SET THE USAGE/HELP   */
	opt->addUsage( "Usage: cocluster [OPTIONS] <input row file> <input column file> <output file>" );
	opt->addUsage( "       -h : print this help");
	opt->addUsage( "       -k N : number of row clusters");
	opt->addUsage( "       -l N : number of column clusters");
	opt->addUsage( "       -m N : number of iterations");
	opt->addUsage( "       -v : verbose");
	opt->addUsage( "       -a : dump temporary assignment at each iteration");
	opt->addUsage( "       -i : compute the loss of information");

	/* 4. SET THE OPTION STRINGS/CHARACTERS */
	opt->setFlag(  "help", 'h' ); /* for help */
	opt->setFlag(  "verbose", 'v');
	opt->setFlag(  "dumpAll", 'a');
	opt->setFlag(  "information", 'i');

	opt->setOption(  "maxiter", 'm' );
	opt->setOption(  "colClusters", 'l');
	opt->setOption(  "rowClusters", 'k');
	opt->setOption(  "numRows", 'n');


	/* 5. PROVIDE THE COMMANDLINE */
	opt->processCommandArgs( argc, argv );

	/* 6. GET THE VALUES */
	/* help */
	if( opt->getFlag( "help" ) || opt->getFlag( 'h' ) ){
		opt->printUsage();		
		exit (-1);
	} else {
		if( opt->getValue( 'm' ) != NULL )
			m = ::atoi(opt->getValue( 'm' ));
		if( opt->getValue( 'l' ) != NULL )
			l = ::atoi(opt->getValue( 'l' ));
		if( opt->getValue( 'k' ) != NULL )
			k = ::atoi(opt->getValue( 'k' ));
		if( opt->getValue( 'n' ) != NULL )
			k = ::atoi(opt->getValue( 'n' ));
		if( opt->getValue( 'v' ) != NULL )
		        verbose = true;
		if( opt->getValue( 'a' ) != NULL )
		        dump_all = true;
		if( opt->getValue( 'i' ) != NULL )
		        information = true;



		/* 7. GET THE ACTUAL ARGUMENTS AFTER THE OPTIONS */
		if (opt->getArgc() < 3) {
			cerr << "Insufficient parameters." << endl;
			opt->printUsage();
			
			exit (-1);
		} else {
			inputRowFile = string(opt->getArgv(0));
			inRow = new ifstream(inputRowFile.c_str());
			inputColumnFile = string(opt->getArgv(1));
			inColumn = new ifstream(inputColumnFile.c_str());
			outputFile = string(opt->getArgv(2));

		}
	}

	/* 7. DONE */
	delete opt;
}

/*
	countRowsColumns analyzes the file named iF counting the number of
	rows and columns contained within the matrix.
	
		input:
			string      if    :	 The name of the input file.
			unsigned    nRows :  A reference for returning the number of Rows;
			unsigned    nCols :  A reference for returning the number of Columns;
			
*/
void countRowsColumns (string iF, unsigned& nRows, unsigned& nCols)
{
	unsigned countedRows = 0;
	unsigned numElements;
	unsigned toSkip;
	unsigned lastColumn;
	float   lastValue;


	ifstream *ifs = new ifstream(iF.c_str());
	
	while (ifs->good()) {
		ifs->read((char*)&numElements, sizeof(unsigned));
		if (ifs->good()) {
			if (numElements > 0) {
				countedRows++;
				toSkip = (numElements - 1) * (sizeof(unsigned)+sizeof(float));
				ifs->seekg(toSkip, ios::cur);
				ifs->read((char*)&lastColumn, sizeof(unsigned));
				ifs->read((char*)&lastValue, sizeof(float));
	
				if (nCols <= lastColumn)
					nCols = lastColumn;
			}
		}
	}
	
	delete ifs;
	
	nRows = countedRows;
	nCols++;
}

void loadMatrix (matrixRow** pRow, matrixRow** pCol, unsigned nRows, unsigned nColumns)
{
	unsigned    numElements;
	matrixEntry *m;

	for (unsigned i = 0; i < nRows; i++) {
		inRow->read((char*)&numElements, sizeof(unsigned));
		
		//		cout << "Row " << i << " has " << numElements << " elements." << endl;
		if (numElements > 0) {
			m = new matrixEntry[numElements];
			inRow->read((char*)m, sizeof(matrixEntry)*numElements);
			pRow[i] = new matrixRow(numElements, m, nColumns);
			delete[] m;
		}
	}
	
	for (unsigned i = 0; i < nColumns; i++) {
		inColumn->read((char*)&numElements, sizeof(unsigned));
		//cout << "Column " << i << " has " << numElements << " elements." << endl;
		if (numElements > 0) {
			m = new matrixEntry[numElements];
			inColumn->read((char*)m, sizeof(matrixEntry)*numElements);
			pCol[i] = new matrixRow(numElements, m, nRows);
			delete[] m;
		}
	}
}

void initializeCoclusters (matrixRow** pxy, matrixRow** pyx, unsigned** cx, unsigned** cy)
{
	// Allocate the cx and cy arrays.
	// cx is an array of length equal to n
	*cx = new unsigned[n];
	
	// cy is an array of lenght equal to nCols
	*cy = new unsigned[nCols];
	
	srand(unsigned(time(NULL)));


	/* 
	   This perform a random initial allocation
	 */
	for (unsigned i = 0; i < n; i++) {

	  //  (*cx)[i] = (i / (n/ k) ); 
	  // use this to get the example from Dillon's paper


	  //(*cx)[i] = (i %k);
	  //round-robin assignment

	  (*cx)[i]=(int) ((k*1.0)*rand()/(RAND_MAX+1.0));
	}
	
	for (unsigned i = 0; i < nCols; i++) {
	  //	  (*cy)[i] = (i / (n /l));
	  //     use this to get the example from Dillon's paper

	  (*cy)[i]=(int) ((l*1.0)*rand()/(RAND_MAX+1.0));

	  //(*cy)[i] = (i %l);
	  //round-robin
	}
}




void save(unsigned* Cx, unsigned* Cy, string name)
{

  string newresname = name+".res";
  ofstream * out = new ofstream(newresname.c_str());

  *out << "<---- Row Clusters ---->" << endl;
  for (unsigned i = 0; i < n; i++) {
    *out << Cx[i] << endl;
  }
  *out << "<---- Column Clusters ---->" << endl;
  for (unsigned i = 0; i < nCols; i++) {
    *out << Cy[i] << endl;
  }
  
  delete out;
}




void save_pcap(matrixRow** pxy, unsigned* Cx, unsigned* Cy, string name)
{

  string newpcapname = name+".pcap";

  ofstream *out = new ofstream(newpcapname.c_str());
	  
  // This is to store the pcapxy
  float  pcapxy[k][l];
  
  // Iterates over pxy rows.
  for (unsigned i = 0; i < k; i++) {
    for (unsigned j = 0; j < l; j++) {
      pcapxy[i][j] = 0;
    }
  }
  for (unsigned i = 0; i < n; i++) {
    for (unsigned j = 0; j < pxy[i]->nColumns(); j++) {
      pcapxy[Cx[i]][Cy[pxy[i]->at(j).columnID]] += pxy[i]->at(j).value;
    }
  }


  for (unsigned i = 0; i < k; i++) {
    for (unsigned j = 0; j < l; j++) {
      *out << pcapxy[i][j] <<" ";
    }
    *out << endl;
  }
  

  delete out;
  
}






bool empty_and_loss(matrixRow** pxy, unsigned* Cx, unsigned* Cy, bool information, double *reported_loss)
{
  //we need to compute pcap(x,y) p(x) p(y) pcap(x) pcap(y)
  float  pcapxy[k][l];
  bool emptycluster = false;

  CREATE(pcapx,k)
  CREATE(pcapy,l)
  CREATE(px,n)
  CREATE(py,nCols)

  
  // We reset pcapxy
  for (unsigned i = 0; i < k; i++) 
    for (unsigned j = 0; j < l; j++) 
      pcapxy[i][j] = 0;




  for (unsigned i = 0; i < n; i++) {
    unsigned rowcl = Cx[i];
    
    for (unsigned j = 0; j < pxy[i]->nColumns(); j++) {

      unsigned colid = pxy[i]->at(j).columnID;
      float value = pxy[i]->at(j).value;
      unsigned colcl = Cy[colid];

      pcapxy[rowcl][colcl] += value;
      pcapx[rowcl]+= value;
      pcapy[colcl] += value;
      px[i] += value;
      py[colid] += value;
    }
  }

  CHECK(pcapx,k)
  CHECK(pcapy,l)


  if (!information)
     return emptycluster;


  float i1 = 0.0;
  float i2 = 0.0;


  for (unsigned i = 0; i < n; i++) {
    unsigned rowcl = Cx[i];
    float pxi = px[i];
    float pcapxrowcl = pcapx[rowcl];

    for (unsigned j = 0; j < pxy[i]->nColumns(); j++) {

      unsigned colid = pxy[i]->at(j).columnID;
      float val = pxy[i]->at(j).value;
      unsigned colcl = Cy[colid];


      i1 += val * log(val / (pxi * py[colid]));
      if (pcapxy[rowcl][colcl]>1e-30)
	i2 += val * log(pcapxy[rowcl][colcl]  /(pcapxrowcl*pcapy[colcl]));
    }
  }

  cout << "LOSS = "<<i1-i2<<endl;

  *reported_loss = i1-i2;
  return emptycluster;
}



void add_noise (unsigned* cx, unsigned* cy, int iterCount)
{

  double threshold = 0.8/iterCount;

 for (unsigned i = 0; i < n; i++) 
   if (toss_coin(threshold))
     cx[i] = (int) ((k*1.0)*rand()/(RAND_MAX+1.0));
						 


 for (unsigned i = 0; i < nCols; i++) 
   if (toss_coin(threshold))
     cy[i] = (int) ((l*1.0)*rand()/(RAND_MAX+1.0));

}




bool termination (matrixRow** pxy, matrixRow** pyx, unsigned* cx, unsigned* cy)
{
	return false;
}




void iterateCoclustering (matrixRow** pxy, matrixRow** pyx, unsigned* cx, unsigned* cy)
{
	// This will store alternatively pcapx and pcapy;
	float* pcap;
	// This is to store the pcapxy
	float  pcapxy[k][l];
	


	// Iterates over pxy rows.
	pcap   = new float[k];
 	for (unsigned i = 0; i < k; i++) {
		for (unsigned j = 0; j < l; j++) {
			pcapxy[i][j] = 0;
		}
	}
	for (unsigned i = 0; i < n; i++) {
		for (unsigned j = 0; j < pxy[i]->nColumns(); j++) {
		  pcapxy[cx[i]][cy[pxy[i]->at(j).columnID]] += pxy[i]->at(j).value;
		}
	}




	//here we compute  Xcap, stored in pcap, summing rows of pcapxy
	for (unsigned i = 0; i < k; i++) {
		pcap[i] = 0;
		for (unsigned j = 0; j < l; j++) {
			pcap[i] += pcapxy[i][j];
		}
	}

	
	float  precomputed[k][l];

 	for (unsigned i = 0; i < k; i++) {
		for (unsigned j = 0; j < l; j++) {
		  //		  bb = pcapxy[j][cy[colid]] / pcap[j];
		  //			  val +=  aa * (log(bb) / log(2.0));		

		  double tmp = pcapxy[i][j]/pcap[i];
		  precomputed[i][j] = log(tmp) / log(2.0);
		}
	}
	

	if (verbose) cout <<"assigning rows ...";

	for (unsigned i = 0; i < n; i++) { //scan x rows
	  
	  completion(i,10000,'.');

	  long int currentArgmax = -1;
	  float    currentMax    = -FLT_MAX;

		for (unsigned j = 0; j < k; j++) { //scan xcap clusters


		        float val = 0.0;
			unsigned nColsCurrentRow = pxy[i]->nColumns();

			for (unsigned ii = 0; ii < nColsCurrentRow; ii++) {
			  //scans elements in the row
			  //accessing the ii-th nonzero column of row i
			  
			  int colid =pxy[i]->at(ii).columnID;
			  float aa = pxy[i]->at(ii).value;
			  //			  float bb = pcapxy[j][cy[colid]] / pcap[j];

			  //			  val +=  aa * (log(bb) / log(2.0));		
			  val += aa * precomputed[j][cy[colid]];
			}


			if (val > currentMax) {
				currentMax = val;
				currentArgmax = j;
			} else 	if (val == currentMax && toss_coin(.5)) {
				currentMax = val;
				currentArgmax = j;
			}
		}
		
		if (currentArgmax==-1)
		  {
		    cerr <<"Unexpected error, probably you have negative entries in your matrix ";
		    cerr <<"Random assignment";
		    
		    (cx)[i]=(int) ((k*1.0)*rand()/(RAND_MAX+1.0));
 		  }
		else
		  cx[i] = currentArgmax;

	}
	delete[] pcap;


	// we rebuild pcap

	pcap   = new float[l];
	for (unsigned i = 0; i < k; i++) {
		for (unsigned j = 0; j < l; j++) {
			pcapxy[i][j] = 0;
		}
	}
	for (unsigned i = 0; i < n; i++) {
		for (unsigned j = 0; j < pxy[i]->nColumns(); j++) {
			pcapxy[cx[i]][cy[pxy[i]->at(j).columnID]] += pxy[i]->at(j).value;
		}
	}


	//here we computer pcap for Ycap, summing columns of pcapxy
	for (unsigned i = 0; i < l; i++) {
		pcap[i] = 0;
		for (unsigned j = 0; j < k; j++) {
			pcap[i] += pcapxy[j][i];
		}
	}


 	for (unsigned i = 0; i < k; i++) {
		for (unsigned j = 0; j < l; j++) {

		  // float bb = pcapxy[cx[rowid]][j] / pcap[j];
		  // val +=  aa * (log(bb) / log(2.0));

		  double tmp = pcapxy[i][j]/pcap[j];
		  precomputed[i][j] = log(tmp) / log(2.0);
		}
	}
	



	if (verbose) cout <<"assigning columns ...";

	for (unsigned i = 0; i < nCols; i++) {//scan y columns
	  completion(i,10000,'.');

		long int currentArgmax = -1;
		float    currentMax    = -FLT_MAX;
		for (unsigned j = 0; j < l; j++) { //scan ycap clusters


			float val = 0.0;
			unsigned nRowsCurrentCol = pyx[i]->nColumns();
			for (unsigned ii = 0; ii < nRowsCurrentCol; ii++) {
			  //scan row elements of columns
			  //  accessing the ii-th nonzero row of column i

				int rowid =pyx[i]->at(ii).columnID;
				float aa = pyx[i]->at(ii).value;
				//				float bb = pcapxy[cx[rowid]][j] / pcap[j];
				//				val +=  aa * (log(bb) / log(2.0));
				val += aa * precomputed[cx[rowid]][j];
			}

			if ((val > currentMax)){
				currentMax = val;
				currentArgmax = j;
			} else 	if (val == currentMax && toss_coin(.5)) {
				currentMax = val;
				currentArgmax = j;
			}

			
		}

		if (currentArgmax==-1)
		  {
		    cerr <<"Unexpected error";
		    cerr <<"Random assignment";
		    
		    (cy)[i]=(int) ((l*1.0)*rand()/(RAND_MAX+1.0));

		  }
		else
		  cy[i] = currentArgmax;
	}
	delete[] pcap;
}











int main (int argc, char** argv)
{
	/* The matrix stored as an array of matrixRow arrays. */
	matrixRow** pxy;
	matrixRow** pyx;
	
	/* The co-clustering is represented as two arrays of arrays of column IDs. */
	unsigned* Cx;
	unsigned* Cy;	
	
	readCommandLine (argc, argv);		
	
	cout << "*****************************************" << endl;
	cout << "  co-clustering algorithm (HPCAPPLE)" << endl;
	cout << "    Implemented by Fabrizio Silvestri" << endl;
	cout << "    and Diego Puppin"<<endl;
	cout << endl;
	cout << endl;
	cout << "  Starting with the following argument" << endl;
	cout << endl;
	cout << "   maxIteration = " << m << endl;
	cout << "   rowClusters = " << k << endl;
	cout << "   columnClusters = " << l << endl;
	if (n > 0) {
		cout << "   numRows = " << n << endl;
	}
	cout << endl;
	cout << "   input row matrix from " << inputRowFile << endl;
	cout << "   input column matrix from " << inputColumnFile << endl;
	cout << "   output into " << outputFile << endl;
	cout << "*****************************************" << endl;
	if (n == 0) {
		cout << " Start counting the number of rows and columns of the matrix. " << endl;
		cout << "Be patient. It may take a while..." << endl;
		
		nCols = 0;
		countRowsColumns (inputRowFile, n, nCols);
		
		cout << "Rows counted. n = " << n << endl;
		cout << "Columns counted. nCols = " << nCols << endl;
	}
	
	pxy = new matrixRow*[n];
	pyx = new matrixRow*[nCols];
	loadMatrix(pxy, pyx, n, nCols);
	
	/* Perform the coclustering. */
	unsigned iterCount = 0;
	initializeCoclusters (pxy, pyx, &Cx, &Cy); // Initialize the coclustering.



	while ((iterCount < m) && (!termination(pxy, pyx, Cx, Cy))) 
	  // Check for termination and continue if the conditions do not hold
	{
	  char buff[200];


	  cout << "Iteration "<< iterCount<<" "<<timestamp(buff)<<"..."<<endl;

	  iterateCoclustering (pxy, pyx, Cx, Cy);


	  if (dump_all)
	    {
	      string newfilename = outputFile+itos(iterCount);
	      
	      if (verbose)
		cout << " saving "<<timestamp(buff)<< "..."<<endl;

	      save(Cx,Cy,newfilename);

	      if (dump_pcap)
		{
		  if (verbose)
		    cout << "saving PCAP "<<timestamp(buff)<<"..."<<endl;

		  save_pcap(pxy, Cx,Cy,newfilename);

		}

	      if (verbose)
		cout << " completed  "<<timestamp(buff)<<endl;


	    }



	  if (information)
	    {
	      double loss;
	      bool empty=empty_and_loss(pxy, Cx,Cy, information, &loss);
	      
	      if (not   (iterCount == m-1))
		if (empty)
		  add_noise(Cx,Cy, iterCount);
	    }

	  iterCount++;

	}




	
	save(Cx,Cy,outputFile);
	if (dump_pcap)
	  save_pcap(pxy, Cx,Cy,outputFile);


	for (unsigned i = 0; i < n; i++) {
		delete pxy[i];
	}
	for (unsigned i = 0; i < nCols; i++) {
		delete pyx[i];
	}
	delete[] pxy;
	delete[] pyx;
	delete inRow;
	delete inColumn;
	return 0;
}
