#include <iostream>

#ifndef _MATRIXROW_
#define _MATRIXROW_

#include "matrixEntry.h"

using namespace std;

class matrixRow {
	public:
		matrixRow (unsigned nColumns, matrixEntry* m, unsigned maxColumns) : nColumns_(nColumns), maxColumns_(maxColumns) {
			this->entries = new matrixEntry[nColumns];
			::memcpy ((void*)this->entries, (const void*)m, sizeof(matrixEntry) * nColumns);
			
			//cout << "Matrix with " << this->nColumns_ << " entries created." << endl;
			//cout << "max columnID is: " << this->entries[nColumns - 1].columnID << endl;
		}
		
		~matrixRow () {
			delete[] this->entries;
		}
		
		unsigned nColumns() {
			return this->nColumns_;
		}
		
		matrixEntry at(unsigned pos) const {
			return this->entries[pos];
		}
		
		void modifyValue (unsigned pos, float value) {
			this->entries[pos].value = value;
		}
		
		float computeSumFromThisSet (unsigned* positions, unsigned nPositions) {
			unsigned  pos        = 0;
			unsigned  toSum      = 0;
			float    sum        = 0.0;
			unsigned  firstPos   = positions[0];
			unsigned  lastPos    = positions[nPositions - 1] + 1;  // + 1 for allowing the use of <
													  																 // (this->entries[pos].columnID < lastPos)
			
			while ((pos < this->nColumns_) && (this->entries[pos].columnID < firstPos)) {
				pos++;
			}

			while ((pos < this->nColumns_) && (this->entries[pos].columnID < lastPos))
			{
				if (this->entries[pos].columnID == positions[toSum]) {
					sum += this->entries[pos].value;
					toSum++;
				}
				pos++;
			}
			
			return sum;
		}
	
		// valueAt returns the value contained at column pos of the matrix
		// This operation is expensive since it has to scan (in the worst case)
		// the entire list of entries;
		float valueAt (unsigned pos) {
			if (pos >= this->maxColumns_) {
				cerr << "Error trying to access an incorrect matrix position: ";
				cerr << pos << " of " << this->maxColumns_ << "." << endl;
				
				exit (-1);
			}
			unsigned i = 0;
			while ((i < this->maxColumns_) && (this->entries[i].columnID < pos))
			{	
				i++;
			}
			if (this->entries[i].columnID == pos)
				return (this->entries[i].value);
			else
				return 0;
		}
		
		// modifyValueAt modify the value contained at column pos inserting v.
		// Expensive operation.
		void modifyValueAt (unsigned pos, float v)
		{
			if (pos >= this->maxColumns_) {
				cerr << "Error trying to access an incorrect matrix position." << endl;
				
				exit (-1);
			}
			int i = 0;
			while (this->entries[i].columnID < pos) i++;
			
			if (this->entries[i].columnID == pos)
				this->entries[i].value = v;
			else {
				cerr << "Error trying to write to an incorrect matrix position." << endl;
			}
		}
		
	private:
		unsigned     nColumns_;
		unsigned     maxColumns_;
		matrixEntry* entries;
};

#endif
